from big_thing_py.utils import *
from big_thing_py.core import MXThing
import functools
import asyncio


class ThingNameRawIDConvertMode:
    ThingNameToRawDeviceID = 0
    RawDeviceIDToThingName = 1


def list_arg_to_str(list_arg: List[Union[int, float, bool, str]]) -> str:
    formatted_list = []
    for x in list_arg:
        if isinstance(x, bool):
            formatted_list.append(str(x))
        elif isinstance(x, float):
            formatted_list.append(f"{x:.4f}")
        elif isinstance(x, (int, str)):
            formatted_list.append(str(x))
        else:
            raise ValueError(f"Unsupported type: {type(x)}")
    return '|'.join(formatted_list)


def generate_enum_list_str(list_str: List[str]) -> str:
    return '|'.join(list_str)


def print_method_info(func: Callable) -> None:
    @functools.wraps(func)
    async def async_wrapper(self: MXThing, *args, **kwargs):
        # MXLOG_DEBUG 호출
        MXLOG_DEBUG(f'{func.__name__} at {self._name} actuate!!!', 'green')
        # 원래 함수 호출
        ret = await func(self, *args, **kwargs)
        return ret

    @functools.wraps(func)
    def sync_wrapper(self: MXThing, *args, **kwargs):
        # MXLOG_DEBUG 호출
        MXLOG_DEBUG(f'{func.__name__} at {self._name} actuate!!!', 'green')
        # 원래 함수 호출
        ret = func(self, *args, **kwargs)
        return ret

    if asyncio.iscoroutinefunction(func):
        async_wrapper.is_decorated = True
        return async_wrapper
    else:
        sync_wrapper.is_decorated = True
        return sync_wrapper


def set_inner_value_buffer(inner_value_buffer: str, target_value: Any = None):
    def decorator(func):
        @functools.wraps(func)
        async def async_wrapper(self: MXThing, *args, **kwargs):
            result = await func(self, *args, **kwargs)

            if not target_value is None:
                setattr(self, inner_value_buffer, target_value)
            else:
                setattr(self, inner_value_buffer, result)
            MXLOG_DEBUG(f'Setting inner value buffer {inner_value_buffer} to {result}.', 'magenta')
            return result

        @functools.wraps(func)
        def sync_wrapper(self: MXThing, *args, **kwargs):
            result = func(self, *args, **kwargs)

            if not target_value is None:
                setattr(self, inner_value_buffer, target_value)
            else:
                setattr(self, inner_value_buffer, result)
            MXLOG_DEBUG(f'Setting inner value buffer {inner_value_buffer} to {result}.', 'magenta')
            return result

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator


def skip_func_execute_if(
    inner_value_buffer: Union[str, List[str]],
    default_return_value: Any = True,
    target_value: Any = None,
    target_argument: Union[str, List[str]] = None,
):
    """
    Decorator to skip the function execution based on the class attribute or function arguments.

    Parameters:
    inner_value_buffer (Union[str, List[str]]): The class attribute name or a list of attribute names.
    target_value (Any): The value to compare with the class attribute(s) or function arguments.
    target_argument_name_list (List[str]): The list of function argument names to compare with the target_value.

    Raises:
    ValueError: If the lengths of inner_value_buffer and target_argument_name_list do not match when inner_value_buffer is a list.
                If target_value is None when inner_value_buffer is not a list.
    """

    def decorator(func):
        @functools.wraps(func)
        def base_wrapper(self, *args, **kwargs) -> bool:
            skip_func_run = True
            inner_value_buffer_instance = getattr(self, inner_value_buffer)

            if isinstance(inner_value_buffer, list):
                if target_argument is None or not isinstance(target_argument, list) or len(inner_value_buffer_instance) != len(target_argument):
                    raise ValueError('The lengths of inner_value_buffer and target_argument_name_list do not match.')

                arg_names = [arg_name for arg_name in func.__code__.co_varnames[1 : func.__code__.co_argcount]]
                target_args = [args[i] for i, _ in enumerate(arg_names)]
                for inner_value, target_arg_value in zip(inner_value_buffer_instance, target_args):
                    if inner_value == target_arg_value:
                        skip_func_run = True
                    else:
                        skip_func_run = False
                        break
            else:
                if not target_value is None and target_argument is None:
                    if inner_value_buffer_instance == target_value:
                        skip_func_run = True
                    else:
                        skip_func_run = False
                elif target_value is None and not target_argument is None:
                    if isinstance(target_argument, list):
                        if not isinstance(inner_value_buffer_instance, list) or len(inner_value_buffer_instance) != len(target_argument):
                            raise ValueError('The lengths of inner_value_buffer and target_argument_name_list do not match.')

                        arg_names = [arg_name for arg_name in func.__code__.co_varnames[1 : func.__code__.co_argcount]]
                        target_args = [args[i] for i, _ in enumerate(arg_names)]
                    elif isinstance(target_argument, str):
                        inner_value_buffer_instance = [inner_value_buffer_instance]
                        target_args = args
                    else:
                        raise ValueError('target_argument_name_list must be a list or a string.')

                    for inner_value, target_arg_value in zip(inner_value_buffer_instance, target_args):
                        if inner_value == target_arg_value:
                            skip_func_run = True
                        else:
                            skip_func_run = False
                            break
                else:
                    raise ValueError('target_value cannot be None when inner_value_buffer is not a list.')

            return skip_func_run
            # if skip_func_run:
            #     MXLOG_DEBUG(f'Skipping function {func.__name__} as the condition is met.', 'yellow')
            #     return default_return_value
            # else:
            #     return func(self, *args, **kwargs)

        @functools.wraps(func)
        async def async_wrapper(self, *args, **kwargs):
            skip_func_run = base_wrapper(self, *args, **kwargs)
            if skip_func_run:
                MXLOG_DEBUG(f'Skipping function {func.__name__} as the condition is met.', 'yellow')
                return default_return_value
            else:
                return await func(self, *args, **kwargs)

        @functools.wraps(func)
        def sync_wrapper(self, *args, **kwargs):
            skip_func_run = base_wrapper(self, *args, **kwargs)
            if skip_func_run:
                MXLOG_DEBUG(f'Skipping function {func.__name__} as the condition is met.', 'yellow')
                return default_return_value
            else:
                return func(self, *args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator


def find_class_in_hierarchy(instance, target_class) -> bool:
    current_class = type(instance)

    while current_class is not object:
        if issubclass(current_class, target_class):
            return True
        current_class = current_class.__base__

    return False


def thing_name_device_id_convert(input: str, mode: ThingNameRawIDConvertMode) -> str:
    # Real device id -> thing name
    if mode == ThingNameRawIDConvertMode.RawDeviceIDToThingName:
        device_id = input

        if ':' in input:
            convert_str = transform_mac_address(device_id, to_colon=False)
        elif '-' in device_id:
            convert_str = input.replace('-', '_')
        else:
            convert_str = device_id
    # thing name -> real device id
    elif mode == ThingNameRawIDConvertMode.ThingNameToRawDeviceID:
        thing_name = input
        category = thing_name.split('__')[0]
        device_id = thing_name.split('__')[1]

        if len(device_id) == 12 and device_id.isupper():
            convert_str = transform_mac_address(device_id, to_colon=True)
        elif '_' in device_id:
            convert_str = device_id.replace('_', '-')
        else:
            convert_str = device_id
    else:
        convert_str = input

    return convert_str



from big_thing_py.core.thing import *


class MXStaffThing(MXThing, metaclass=ABCMeta):

    def __init__(
        self,
        name: str,
        nick_name: str,
        category: DeviceCategory,
        desc: str,
        version: str,
        service_list: List[MXService],
        alive_cycle: float,
        is_parallel: bool,
        is_matter: bool,
    ):
        super().__init__(
            name=name,
            nick_name=nick_name,
            desc=desc,
            category=category,
            version=version,
            service_list=service_list,
            alive_cycle=alive_cycle,
            is_super=False,
            is_parallel=is_parallel,
            is_builtin=False,
            is_manager=False,
            is_matter=is_matter,
        )
        self._is_alive = False

        self._default_tag_list: List[MXTag] = [
            MXTag(tag if check_valid_identifier(tag) == MXErrorCode.NO_ERROR else convert_to_valid_string(tag))
            for tag in [
                self._name,
                self._nick_name,
            ]
            if tag
        ]

    def __eq__(self, o: 'MXStaffThing') -> bool:
        instance_check = isinstance(o, MXStaffThing)
        staff_thing_name_check = o._name == self._name

        # return super().__eq__(o) and instance_check and staff_thing_id_check
        return instance_check and staff_thing_name_check

    def add_staff_service(self, service_list: List[MXService]) -> None:
        for staff_service in service_list:
            self.add_service(staff_service)

    @abstractmethod
    def staff_dict(self):
        pass

    @abstractmethod
    def setup_service_list(self) -> None:
        pass

    @abstractmethod
    async def setup(self) -> bool:
        pass

    @abstractmethod
    async def wrapup(self) -> bool:
        pass

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def is_alive(self) -> bool:
        return self._is_alive

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @is_alive.setter
    def is_alive(self, is_alive: bool) -> None:
        self._is_alive = is_alive



from big_thing_py.common import *

from termcolor import colored, cprint

import logging
import os
import time
import re
from enum import Enum
from datetime import datetime


class MicrosecondFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = datetime.now().strftime(datefmt)
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s,%03d" % (t, record.msecs)  # 마이크로초까지 출력
        return s


class MicrosecondFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = datetime.now().strftime(datefmt)
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s,%03d" % (t, record.msecs)  # 마이크로초까지 출력
        return s


class MXLogger:
    class LoggingMode(Enum):
        ALL = 0
        FILE = 1
        CONSOLE = 2
        OFF = 3

    class PrintMode(Enum):
        DEBUG = 0
        INFO = 1
        WARN = 2
        ERROR = 3
        CRITICAL = 4

    def __init__(
        self,
        file_name: str = f'mqtt_message_{time.strftime("%Y%m%d%H%M", time.localtime(time.time()))}.log',
        file_path: str = f'./log',
        logging_mode: LoggingMode = LoggingMode.ALL,
    ) -> None:
        self._file_name = file_name
        self._file_path = file_path
        self._logging_mode = logging_mode

        self._console_logger = None
        self._file_logger = None

    def start(self):
        formatter = MicrosecondFormatter('[%(asctime)s] %(message)s', datefmt='%Y/%m/%d %H:%M:%S.%f')
        # formatter = logging.Formatter('[%(asctime)s.%(msecs)d] %(message)s', datefmt='%Y/%m/%d %H:%M:%S')
        level_list = [logging.DEBUG, logging.INFO, logging.WARN, logging.ERROR, logging.CRITICAL]

        if self._logging_mode == self.LoggingMode.ALL:
            os.makedirs('/'.join(self._file_path.rstrip('/').split('/')), exist_ok=True)

            file_logger = logging.getLogger('file_logger')
            file_logger.setLevel(logging.DEBUG)
            file_handler_list = [
                logging.FileHandler(filename='/'.join([self._file_path, self._file_name]), mode='a', encoding='utf-8') for _ in range(5)
            ]

            for level, file_handler in zip(level_list, file_handler_list):
                file_handler.setLevel(level)
                file_handler.setFormatter(formatter)
                file_logger.addHandler(file_handler)

            self._file_logger = file_logger

            console_logger = logging.getLogger('console_logger')
            console_logger.setLevel(logging.DEBUG)

            console_handler_list = [logging.StreamHandler() for _ in range(5)]

            for level, console_handler in zip(level_list, console_handler_list):
                console_handler.setLevel(level)
                console_handler.setFormatter(formatter)
                console_logger.addHandler(console_handler)

            self._console_logger = console_logger
        elif self._logging_mode == self.LoggingMode.FILE:
            os.makedirs('/'.join(self._file_path.rstrip('/').split('/')), exist_ok=True)

            file_logger = logging.getLogger('file_logger')
            file_logger.setLevel(logging.DEBUG)
            file_handler_list = [logging.FileHandler(filename='/'.join([self._file_path, self._file_name]), mode='a') for _ in range(5)]

            for level, file_handler in zip(level_list, file_handler_list):
                file_handler.setLevel(level)
                file_handler.setFormatter(formatter)
                file_logger.addHandler(file_handler)

            self._file_logger = file_logger
        elif self._logging_mode == self.LoggingMode.CONSOLE:
            console_logger = logging.getLogger('console_logger')
            console_logger.setLevel(logging.DEBUG)

            console_handler_list = [logging.StreamHandler() for _ in range(5)]

            for level, console_handler in zip(level_list, console_handler_list):
                console_handler.setLevel(level)
                console_handler.setFormatter(formatter)
                console_logger.addHandler(console_handler)

            self._console_logger = console_logger
        else:
            pass

    def _remove_color(self, msg: str) -> str:
        ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')
        msg = ansi_escape.sub('', msg)
        return msg

    def select_by_logger_name(self, logger_name: str, logging_func: Callable, msg: List[str], color: str):
        if logger_name == 'console_logger':
            logging_func(colored(msg, color))
        elif logger_name == 'file_logger':
            logging_func(self._remove_color(msg))

    def select_by_print_mode(self, logger: logging.Logger, msg: List[str], color: str, print_mode: PrintMode = PrintMode.DEBUG):
        if print_mode == self.PrintMode.DEBUG:
            self.select_by_logger_name(logger.name, logger.debug, msg, color)
        elif print_mode == self.PrintMode.INFO:
            self.select_by_logger_name(logger.name, logger.info, msg, color)
        elif print_mode == self.PrintMode.WARN:
            self.select_by_logger_name(logger.name, logger.warning, msg, color)
        elif print_mode == self.PrintMode.ERROR:
            self.select_by_logger_name(logger.name, logger.error, msg, color)
        elif print_mode == self.PrintMode.CRITICAL:
            self.select_by_logger_name(logger.name, logger.critical, msg, color)
        else:
            cprint(f'[MXLOG_DEBUG] not supported print mode...', 'red')

    def print(self, msg: List[str], color: str = None, mode: PrintMode = PrintMode.DEBUG):
        try:
            if self._logging_mode == self.LoggingMode.ALL:
                self.select_by_print_mode(self._console_logger, msg, color, mode)
                self.select_by_print_mode(self._file_logger, msg, color, mode)
            elif self._logging_mode == self.LoggingMode.FILE:
                self.select_by_print_mode(self._file_logger, msg, color, mode)
            elif self._logging_mode == self.LoggingMode.CONSOLE:
                self.select_by_print_mode(self._console_logger, msg, color, mode)
            elif self._logging_mode == self.LoggingMode.OFF:
                pass
            else:
                raise Exception(f'[MXLOG_DEBUG] Not supported logging mode ')
        except Exception as e:
            print(f'[MXLOG_DEBUG] Unknown exception error : {str(e)}')
            pass


base_logger: MXLogger = None


def START_LOGGER(whole_log_path: str = None, logging_mode: MXLogger.LoggingMode = MXLogger.LoggingMode.ALL, append_time_to_file_name: bool = False):
    global base_logger

    append_log_file_name = f'_{time.strftime("%Y%m%d%H%M", time.localtime(time.time()))}' if append_time_to_file_name else ''
    if not whole_log_path:
        whole_log_path = f'./log/mqtt_message{append_log_file_name}.log'
    else:
        whole_log_path = f'./{whole_log_path.rsplit(".", 1)[0]}{append_log_file_name}.log'
    os.makedirs(os.path.dirname(whole_log_path), exist_ok=True)
    log_name = os.path.basename(whole_log_path)
    log_path = os.path.dirname(whole_log_path)

    if base_logger is None:
        base_logger = MXLogger(file_name=log_name, file_path=log_path, logging_mode=logging_mode)
        base_logger.start()
    # cprint('logger is started!', 'green')


def MXLOG_DEBUG(msg: List[str], color: str = None, mode: MXLogger.PrintMode = MXLogger.PrintMode.DEBUG):
    global base_logger
    if base_logger is None:
        START_LOGGER()

    base_logger.print(msg, color, mode)


def MXLOG_WARN(msg: List[str], color: str = 'yellow', mode: MXLogger.PrintMode = MXLogger.PrintMode.DEBUG):
    MXLOG_DEBUG(msg, color, mode)


def MXLOG_ERROR(msg: List[str], color: str = 'red', mode: MXLogger.PrintMode = MXLogger.PrintMode.DEBUG):
    MXLOG_DEBUG(msg, color, mode)


if __name__ == '__main__':
    # START_LOGGER(logging_mode=MXLogger.LoggingMode.ALL)
    START_LOGGER(logging_mode=MXLogger.LoggingMode.OFF)
    # START_LOGGER(logging_mode=MXLogger.LoggingMode.CONSOLE)
    MXLOG_DEBUG('test', 'red')
    MXLOG_DEBUG('test')
    # logger.print(f'INFO test', color='red', mode=MXLogger.PrintMode.INFO)
    # logger.print(f'WARN test', color='red', mode=MXLogger.PrintMode.WARN)
    # logger.print(f'ERROR test', color='red', mode=MXLogger.PrintMode.ERROR)
    # logger.print(f'CRITICAL test', color='red',
    #              mode=MXLogger.PrintMode.CRITICAL)
